/*
 * Copyright 2009 bufferings[at]gmail.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package bufferings.ktr.i18n.server;

import java.lang.reflect.Constructor;
import java.security.ProtectionDomain;
import java.util.Locale;

import javassist.CannotCompileException;
import javassist.ClassClassPath;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.NotFoundException;
import bufferings.ktr.i18n.client.KtrI18NException;

/**
 * The generator template for the resource access class.
 * 
 * @author bufferings[at]gmail.com
 */
public abstract class GeneratorTemplate {

  /** class pool of javassist */
  protected ClassPool pool;

  /** default constructor */
  public GeneratorTemplate() {
    pool = ClassPool.getDefault();
    pool.insertClassPath(new ClassClassPath(GeneratorTemplate.class));
  }

  /**
   * Generates and instantiates a class from the interface name and the locale
   * information.
   * 
   * @param <T>
   *          a type of the return class
   * @param interfaceName
   *          the interface name (contains package name)
   * @param language
   *          lowercase two-letter ISO-639 code.
   * @param country
   *          uppercase two-letter ISO-3166 code.
   * @return the new instance of the generated implementation
   * @throws KtrI18NException
   *           if the generation or the instantiation fails
   */
  public <T> T generate(String interfaceName, String language, String country) {
    try {
      String implClassName = getImplClassName(interfaceName);

      Class<T> clazz = loadClass(implClassName);
      if (clazz == null) {
        String resourceName = getResourceName(interfaceName);
        CtClass interfaceCtc = getCtClass(interfaceName);

        CtClass implCtc = makeImplCtClass(implClassName, interfaceCtc);
        addResourceBundleField(implCtc);
        addConstructor(implCtc, resourceName);
        addMethods(implCtc, interfaceCtc);

        clazz = getClassFromCtClass(implCtc);

        implCtc.detach();
      }

      return getInstanceFromClass(clazz, language, country);
    } catch (KtrI18NException e) {
      throw e;
    } catch (Exception e) {
      throw new KtrI18NException("Generation failed", e);
    }
  }

  /**
   * Gets a implementation class name from the interface name.
   * 
   * @param interfaceName
   *          the interface name
   * @return the implementation class name
   */
  protected String getImplClassName(String interfaceName) {
    return interfaceName + "Impl";
  }

  /**
   * Gets a resource file name from the interface name.
   * 
   * @param interfaceName
   *          the interface name
   * @return resource file name
   */
  protected String getResourceName(String interfaceName) {
    return interfaceName;
  }

  /**
   * Loads a class using the class loader of the current thread context.
   * 
   * @param className
   *          the class name
   * @return the loaded class. If not found, returns null.
   */
  @SuppressWarnings("unchecked")
  protected <T> Class<T> loadClass(String className) {
    try {
      return (Class<T>) Thread
        .currentThread()
        .getContextClassLoader()
        .loadClass(className);
    } catch (ClassNotFoundException e1) {
      return null;
    }
  }

  /**
   * Gets a class definition of javassist.
   * 
   * @param className
   *          the class name
   * @return the class definition
   * @throws KtrI18NException
   *           if it fails to get the class definition
   */
  protected CtClass getCtClass(String className) {
    try {
      return pool.get(className);
    } catch (NotFoundException e) {
      throw new KtrI18NException("Getting the CtClass failed(className="
        + className
        + ").", e);
    }
  }

  /**
   * Makes a implementation class definition with the interface.
   * 
   * @param className
   *          the class name
   * @param interfaceCtc
   *          the definition of the interface to attach to the class
   * @return the implementation class definition
   */
  protected CtClass makeImplCtClass(String className, CtClass interfaceCtc) {
    CtClass implCtc = pool.makeClass(className);
    implCtc.addInterface(interfaceCtc);
    return implCtc;
  }

  /**
   * Adds a field of the resource bundle to the class definition.
   * 
   * @param ctc
   *          the class definition
   * @throws KtrI18NException
   *           if it fails to add the field
   */
  protected void addResourceBundleField(CtClass ctc) {
    String body = "protected java.util.ResourceBundle rb;";

    try {
      CtField f = CtField.make(body, ctc);
      ctc.addField(f);
    } catch (CannotCompileException e) {
      throw new KtrI18NException("Adding the ResourceBundle field failed.", e);
    }
  }

  /**
   * Adds a constructor to the class definition.
   * 
   * @param ctc
   *          the class definition
   * @param resourceName
   *          resource name
   * @throws KtrI18NException
   *           if it fails to get the {@code java.util.Locale} definition
   * @throws KtrI18NException
   *           if it fails to add the constructor
   */
  protected void addConstructor(CtClass ctc, String resourceName) {
    CtClass localeCtc = getCtClass("java.util.Locale");
    CtClass[] parameters = new CtClass[] { localeCtc };

    String body =
      String.format(
        "rb = java.util.ResourceBundle.getBundle(\"%s\",$1);",
        resourceName);

    try {
      CtConstructor c = new CtConstructor(parameters, ctc);
      c.setBody(body);
      ctc.addConstructor(c);
    } catch (CannotCompileException e) {
      throw new KtrI18NException("Adding the constructor failed.", e);
    }
  }

  /**
   * Gets a class from the class definition.
   * 
   * @param ctc
   *          the class definition
   * @return the class
   * @throws KtrI18NException
   *           if it fails to get the class
   */
  @SuppressWarnings("unchecked")
  protected <T> Class<T> getClassFromCtClass(CtClass ctc) {
    ClassLoader cl = Thread.currentThread().getContextClassLoader();
    ProtectionDomain pd = KtrI18NCreator.class.getProtectionDomain();
    Class<T> clazz;
    try {
      clazz = (Class<T>) ctc.toClass(cl, pd);
    } catch (CannotCompileException e) {
      throw new KtrI18NException("Getting the class failed(className="
        + ctc.getName()
        + ").", e);
    }
    return clazz;
  }

  /**
   * Instantiates a class.
   * 
   * @param <T>
   *          the class type
   * @param clazz
   *          the class
   * @param language
   *          lowercase two-letter ISO-639 code.
   * @param country
   *          uppercase two-letter ISO-3166 code.
   * @return the instance of the class
   * @throws KtrI18NException
   *           if it fails to instantiate the class
   */
  protected <T> T getInstanceFromClass(Class<T> clazz, String language,
      String country) {
    Locale locale = new Locale(language, country);

    try {
      Constructor<T> constructor = clazz.getConstructor(Locale.class);
      return constructor.newInstance(locale);
    } catch (Exception e) {
      throw new KtrI18NException("Getting the instance failed(className="
        + clazz.getName()
        + ").", e);
    }
  }

  /**
   * Adds method definitions from the interface definition.
   * 
   * @param implCtc
   *          the class definiion of the implementation to add the method
   * @param interfaceCtc
   *          the class definiion of the interface
   */
  protected void addMethods(CtClass implCtc, CtClass interfaceCtc) {
    CtMethod[] interfaceCtMethods = interfaceCtc.getDeclaredMethods();
    for (CtMethod interfaceCtMethod : interfaceCtMethods) {
      addMethod(implCtc, interfaceCtMethod);
    }
  }

  /**
   * Adds a method definition from the interface definition.
   * 
   * @param implCtc
   *          the definition of the implementation class.
   * @param interfaceCtMethod
   *          the definition of the interface method
   * @throws KtrI18NException
   *           if it fails to add method.
   */
  protected void addMethod(CtClass implCtc, CtMethod interfaceCtMethod) {
    String methodName = interfaceCtMethod.getName();
    CtClass returnType = getMethodReturnType(interfaceCtMethod);
    CtClass[] paramTypes = getMethodParameterTypes(interfaceCtMethod);

    CtMethod implCtMethod =
      new CtMethod(returnType, methodName, paramTypes, implCtc);
    setImplMethodBody(implCtMethod);

    try {
      implCtc.addMethod(implCtMethod);
    } catch (CannotCompileException e) {
      throw new KtrI18NException("Adding the method failed(methodName="
        + methodName
        + ").", e);
    }
  }

  /**
   * Gets the return type class definition from the method definition.
   * 
   * @param ctMethod
   *          the method definition
   * @return the return type class definition of the method definition.
   * @throws KtrI18NException
   *           If it fails to get the return type class definition.
   */
  protected CtClass getMethodReturnType(CtMethod ctMethod) {
    try {
      return ctMethod.getReturnType();
    } catch (NotFoundException e) {
      throw new KtrI18NException(
        "Getting the return type of the method failed(methodName="
          + ctMethod.getName()
          + ").",
        e);
    }
  }

  /**
   * Gets method parameter type class definitions from the method definition.
   * 
   * @param ctMethod
   *          the method definition
   * @return array of parameter class definitions of the method
   * @throws KtrI18NException
   *           If it fails to get method parameter type class definitions.
   */
  protected CtClass[] getMethodParameterTypes(CtMethod ctMethod) {
    try {
      return ctMethod.getParameterTypes();
    } catch (NotFoundException e) {
      throw new KtrI18NException(
        "Adding the parameters of the method failed(methodName="
          + ctMethod.getName()
          + ").",
        e);
    }
  }

  /**
   * Sets a method body to the method definition.
   * 
   * @param implCtMethod
   *          the method definition
   * @throws KtrI18NException
   *           If it fails to set a method body.
   */
  protected abstract void setImplMethodBody(CtMethod implCtMethod);
}
